# global modules
import json
import unittest

import numpy as np
import torch
from lifelines import WeibullAFTFitter
from lifelines.datasets import load_gbsg2, load_lung

# Local modules
from torchsurv.loss.weibull import neg_log_likelihood as weibull

# Load the benchmark cox log likelihoods from R
with open("tests/benchmark_data/benchmark_weibull.json", "r") as file:
    benchmark_weibull_logliks = json.load(file)

# set seed for reproducibility
torch.manual_seed(42)


class TestWeibullSurvivalLoss(unittest.TestCase):
    """
    List of packages compared
        - survival (R)
        - lifelines (python)
    """

    # random data and parameters
    N = 32
    log_params = torch.randn(N, 2)
    y = torch.randint(low=0, high=2, size=(N, 1)).bool()
    t = torch.randint(low=1, high=100, size=(N, 1))

    # prepare data
    lung = load_lung()
    lung["sex"] = (lung["sex"] == 1).astype(float)
    lung["age"] = (lung["age"] - lung["age"].mean()) / lung["age"].std()

    gbsg = load_gbsg2()
    gbsg["age"] = (gbsg["age"] - gbsg["age"].mean()) / gbsg["age"].std()
    gbsg["tsize"] = (gbsg["tsize"] - gbsg["tsize"].mean()) / gbsg["tsize"].std()

    def test_y_tensor(self):
        y_np_array = np.random.randint(0, 1 + 1, size=(self.N, 1), dtype="bool")
        self.assertRaises(TypeError, weibull, self.log_params, y_np_array, self.t)

    def test_t_tensor(self):
        t_np_array = np.random.randint(0, 100, size=(self.N, 1))
        self.assertRaises(TypeError, weibull, self.log_params, self.y, t_np_array)

    def test_log_params_tensor(self):
        log_params_np_array = np.random.randn(self.N, 2)
        self.assertRaises(TypeError, weibull, log_params_np_array, self.y, self.t)

    def test_nrow_log_params(self):
        log_params_wrong_nrow = torch.randn(self.N + 1, 2)
        self.assertRaises(ValueError, weibull, log_params_wrong_nrow, self.y, self.t)

    def test_len_data(self):
        t_wrong_len = torch.randint(low=1, high=100, size=(self.N + 1, 1))
        self.assertRaises(ValueError, weibull, self.log_params, self.y, t_wrong_len)

    def test_positive_t(self):
        t_negative = torch.randint(low=-100, high=100, size=(self.N, 1))
        self.assertRaises(ValueError, weibull, self.log_params, self.y, t_negative)

    def test_boolean_y(self):
        y_non_boolean = torch.randint(low=0, high=3, size=(self.N, 1))
        self.assertRaises(ValueError, weibull, self.log_params, y_non_boolean, self.t)

    def test_log_likelihood_1_param(self):
        """test weibull log likelihood with only 1 param (log scale) on lung and gbsg data"""
        for benchmark_weibull_loglik in benchmark_weibull_logliks:
            log_lik = -weibull(
                torch.tensor(
                    benchmark_weibull_loglik["log_shape"], dtype=torch.float32
                ).squeeze(0),
                torch.tensor(benchmark_weibull_loglik["status"]).bool(),
                torch.tensor(benchmark_weibull_loglik["time"], dtype=torch.float32),
                reduction="sum",
            )

            log_lik_survival = benchmark_weibull_loglik["log_likelihood"]

            self.assertTrue(
                np.allclose(
                    log_lik.numpy(),
                    np.array(log_lik_survival),
                    rtol=1e-5,
                    atol=1e-8,
                )
            )

    def test_log_likelihood_2_params_lung(self):
        """test value of weibull log likelihood with 2 params (log scale and log shape) on lung data"""

        event = torch.tensor(self.lung["status"]).bool()
        time = torch.tensor(self.lung["time"], dtype=torch.float32)

        # intercept only
        wbf = WeibullAFTFitter()
        wbf.fit(self.lung[["time", "status"]], duration_col="time", event_col="status")
        log_params = np.ones((len(event), 1)) * wbf.summary.loc[:, "coef"].values
        log_likelihood_lifelines = wbf.log_likelihood_
        log_likelihood = -weibull(
            torch.tensor(log_params, dtype=torch.float32),
            event,
            time,
            reduction="sum",
        )
        self.assertTrue(
            np.allclose(
                log_likelihood.numpy(),
                np.array(log_likelihood_lifelines),
                rtol=1e-5,
                atol=1e-8,
            )
        )

        # one covariate
        wbf = WeibullAFTFitter()
        wbf.fit(
            self.lung[["time", "status", "age"]],
            duration_col="time",
            event_col="status",
        )
        log_scale = (
            wbf.summary.loc[:, "coef"].lambda_.Intercept
            + np.array(self.lung["age"]) * wbf.summary.loc[:, "coef"].lambda_.age
        )
        log_shape = (
            np.ones((self.lung.shape[0],)) * wbf.summary.loc[:, "coef"].rho_.Intercept
        )
        log_params = np.column_stack((log_scale, log_shape))
        log_likelihood_lifelines = wbf.log_likelihood_
        log_likelihood = -weibull(
            torch.tensor(log_params, dtype=torch.float32),
            event,
            time,
            reduction="sum",
        )
        self.assertTrue(
            np.allclose(
                log_likelihood.numpy(),
                np.array(log_likelihood_lifelines),
                rtol=1e-5,
                atol=1e-8,
            )
        )

        # two covariates
        wbf = WeibullAFTFitter()
        wbf.fit(
            self.lung[["time", "status", "age", "sex"]],
            duration_col="time",
            event_col="status",
        )
        log_scale = (
            wbf.summary.loc[:, "coef"].lambda_.Intercept
            + np.array(self.lung["age"]) * wbf.summary.loc[:, "coef"].lambda_.age
            + np.array(self.lung["sex"]) * wbf.summary.loc[:, "coef"].lambda_.sex
        )
        log_shape = (
            np.ones((self.lung.shape[0],)) * wbf.summary.loc[:, "coef"].rho_.Intercept
        )
        log_params = np.column_stack((log_scale, log_shape))
        log_likelihood_lifelines = wbf.log_likelihood_
        log_likelihood = -weibull(
            torch.tensor(log_params, dtype=torch.float32),
            event,
            time,
            reduction="sum",
        )
        self.assertTrue(
            np.allclose(
                log_likelihood.numpy(),
                np.array(log_likelihood_lifelines),
                rtol=1e-5,
                atol=1e-8,
            )
        )

    def test_log_likelihood_2_params(self):
        """test value of weibull log likelihood with 2 params (log scale and log shape) on gbsg data"""

        event = torch.tensor(self.gbsg["cens"]).bool()
        time = torch.tensor(self.gbsg["time"], dtype=torch.float32)

        # intercept only
        wbf = WeibullAFTFitter()
        wbf.fit(self.gbsg[["time", "cens"]], duration_col="time", event_col="cens")
        log_params = np.ones((len(event), 1)) * wbf.summary.loc[:, "coef"].values
        log_likelihood_lifelines = wbf.log_likelihood_
        log_likelihood = -weibull(
            torch.tensor(log_params, dtype=torch.float32),
            event,
            time,
            reduction="sum",
        )
        self.assertTrue(
            np.allclose(
                log_likelihood.numpy(),
                np.array(log_likelihood_lifelines),
                rtol=1e-5,
                atol=1e-8,
            )
        )

        # one covariate
        wbf = WeibullAFTFitter()
        wbf.fit(
            self.gbsg[["time", "cens", "age"]],
            duration_col="time",
            event_col="cens",
        )
        log_scale = (
            wbf.summary.loc[:, "coef"].lambda_.Intercept
            + np.array(self.gbsg["age"]) * wbf.summary.loc[:, "coef"].lambda_.age
        )
        log_shape = (
            np.ones((self.gbsg.shape[0],)) * wbf.summary.loc[:, "coef"].rho_.Intercept
        )
        log_params = np.column_stack((log_scale, log_shape))
        log_likelihood_lifelines = wbf.log_likelihood_
        log_likelihood = -weibull(
            torch.tensor(log_params, dtype=torch.float32),
            event,
            time,
            reduction="sum",
        )
        self.assertTrue(
            np.allclose(
                log_likelihood.numpy(),
                np.array(log_likelihood_lifelines),
                rtol=1e-5,
                atol=1e-8,
            )
        )

        # two covariates
        wbf = WeibullAFTFitter()
        wbf.fit(
            self.gbsg[["time", "cens", "age", "tsize"]],
            duration_col="time",
            event_col="cens",
        )
        log_scale = (
            wbf.summary.loc[:, "coef"].lambda_.Intercept
            + np.array(self.gbsg["age"]) * wbf.summary.loc[:, "coef"].lambda_.age
            + np.array(self.gbsg["tsize"]) * wbf.summary.loc[:, "coef"].lambda_.tsize
        )
        log_shape = (
            np.ones((self.gbsg.shape[0],)) * wbf.summary.loc[:, "coef"].rho_.Intercept
        )
        log_params = np.column_stack((log_scale, log_shape))
        log_likelihood_lifelines = wbf.log_likelihood_
        log_likelihood = -weibull(
            torch.tensor(log_params, dtype=torch.float32),
            event,
            time,
            reduction="sum",
        )
        self.assertTrue(
            np.allclose(
                log_likelihood.numpy(),
                np.array(log_likelihood_lifelines),
                rtol=1e-5,
                atol=1e-8,
            )
        )


if __name__ == "__main__":
    unittest.main()



import torch


def validate_survival_data(event, time):
    """Perform format and validity checks for survival data.

    Args:
        event (torch.Tensor, boolean):
            Event indicator of size n_samples (= True if event occured).
        time (torch.Tensor, float):
            Event or censoring time of size n_samples.

    Raises:
        TypeError: If ``event`` or ``time`` are not tensors.
        ValueError: If ``event`` is not boolean.
        ValueError: If ``event`` and ``time`` are not of the same length.
        ValueError: If all ``event`` are False.
        ValueError: If any ``time`` are negative.
    """
    if not torch.is_tensor(event) or not torch.is_tensor(time):
        raise TypeError("Inputs 'event' and 'time' should be tensors")

    if not event.dtype == torch.bool:
        raise ValueError("Input 'event' should be of boolean type.")

    if not torch.is_floating_point(time):
        raise ValueError("Input 'time' should be of float type.")

    if len(event) != len(time):
        raise ValueError(
            "Dimension mismatch: Incompatible length between inputs 'time' and 'event'."
        )

    if torch.sum(event) <= 0:
        raise ValueError("All samples are censored.")

    if torch.any(time < 0.0):
        raise ValueError("Input 'time' should be non-negative.")


def validate_evaluation_time(new_time, time, within_follow_up=True):
    """Perform format and validity checks for evaluation time.

    Args:
        new_time (torch.Tensor, optional):
            Time points for metric computation of size n_times.
        time (torch.Tensor, float):
            Event or censoring time of size n_samples.
        within_follow_up (bool, optional):
            Whether values of ``new_time`` must be within values in ``time``.
            Defaults to True.

    Raises:
        ValueError: If ``new_time`` contains duplicate values.
        ValueError: If ``new_time`` is not sorted.
        TypeError: If ``new_time`` is not a tensor.
        ValueError: If ``new_time`` is not of floating-point type.
        ValueError:
            If ``new_time`` is not within the range of follow-up in ``time``.
            Assessed only if ``within_follow_up`` is True.
    """
    new_time_sorted, indices = torch.unique(new_time, return_inverse=True)

    if len(new_time_sorted) != len(new_time):
        raise ValueError("'Value error: Input 'new_time' should contain unique values.")

    if not torch.all(indices == torch.arange(len(new_time))):
        raise ValueError(
            "Value error: Input 'new_time' should be sorted from the smallest time to the largest."
        )

    if not torch.is_tensor(new_time):
        raise TypeError("Type error: Input 'new_time' should be a tensor.")

    if not torch.is_floating_point(new_time):
        raise ValueError(
            "Value error: Input 'new_time' should be of floating-point type."
        )

    if within_follow_up:
        if new_time.max() >= time.max() or new_time.min() < time.min():
            raise ValueError(
                "Value error: All new_time must be within follow-up time of test data: [{}; {}[".format(
                    time.min(), time.max()
                )
            )


def validate_estimate(estimate, time, new_time=None) -> None:
    """Perform format and validity checks for estimate.

    Args:
        estimate (torch.Tensor):
            Estimates of shape = (n_samples,) or (n_samples, n_times).
        time (torch.Tensor, float):
            Time of event or censoring of size n_samples.
        new_time (torch.Tensor, optional):
            Time points for metric computation of size n_times.

    Raises:
        TypeError: If ``estimate`` is not a tensor.
        ValueError: If ``estimate`` has more than 2 dimensions.
        ValueError: If number of rows of ``estimate`` is not n_samples.
        ValueError:
            If number of columns of ``estimate`` is not n_times.
            Assessed only if ``new_time`` is not None.
    """
    if not torch.is_tensor(estimate):
        raise TypeError("Type error: Input 'estimate' should be a tensor.")

    if estimate.ndim > 2:
        raise ValueError("Value error: Input 'estimate' should have 1 or 2 dimensions.")

    if estimate.shape[0] != time.shape[0]:
        raise ValueError(
            "Dimension mismatch: Inconsistent number of samples between inputs 'time' and 'estimate'."
        )

    if not new_time is None:
        if estimate.ndim == 2 and estimate.shape[1] != new_time.shape[0]:
            raise ValueError(
                "Dimension mismatch: Expected  inputs 'estimate' with {} columns, but got {}".format(
                    new_time.shape[0], estimate.shape[1]
                )
            )



import copy
import sys
from typing import Optional

import torch
from scipy import stats

from ..tools import validate_inputs


class BrierScore:
    r"""Compute the Brier Score for survival models."""

    def __init__(self, checks: bool = True):
        """Initialize a BrierScore for survival class model evaluation.

        Args:
            checks (bool):
                Whether to perform input format checks.
                Enabling checks can help catch potential issues in the input data.
                Defaults to True.

        Examples:
            >>> _ = torch.manual_seed(52)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.rand((n,len(time)))
            >>> brier_score = BrierScore()
            >>> brier_score(estimate, event, time)
            tensor([0.2463, 0.2740, 0.3899, 0.1964, 0.3608, 0.2821, 0.1932, 0.2978, 0.1950,
                    0.1668])
            >>> brier_score.integral() # integrated brier score
            tensor(0.2862)
            >>> brier_score.confidence_interval() # default: parametric, two-sided
            tensor([[0.1061, 0.0604, 0.2360, 0.0533, 0.1252, 0.0795, 0.0000, 0.1512, 0.0381,
                     0.0051],
                    [0.3866, 0.4876, 0.5437, 0.3394, 0.5965, 0.4847, 0.4137, 0.4443, 0.3520,
                     0.3285]])
            >>> brier_score.p_value() # default: bootstrap permutation test, two-sided
            tensor([1.0000, 0.7860, 1.0000, 0.3840, 1.0000, 1.0000, 0.3840, 1.0000, 0.7000,
                    0.2380])
        """
        self.checks = checks

        # init instate attributes
        self.order_time = None
        self.time = None
        self.event = None
        self.weight = None
        self.new_time = None
        self.weight_new_time = None
        self.estimate = None
        self.brier_score = None
        self.residuals = None

    def __call__(
        self,
        estimate: torch.Tensor,
        event: torch.Tensor,
        time: torch.Tensor,
        new_time: Optional[torch.Tensor] = None,
        weight: Optional[torch.Tensor] = None,
        weight_new_time: Optional[torch.Tensor] = None,
        instate: bool = True,
    ) -> torch.Tensor:
        r"""Compute the Brier Score.

        Args:
            estimate (torch.Tensor):
                Estimated probability of remaining event-free (i.e., survival function).
                Can be of shape = (n_samples, n_samples) if subject-specific survival is evaluated at ``time``,
                or of shape = (n_samples, n_times) if subject-specific survival is evaluated at ``new_time``.
            event (torch.Tensor, boolean):
                Event indicator of size n_samples (= True if event occured)
            time (torch.Tensor, float):
                Time-to-event or censoring of size n_samples.
            new_time (torch.Tensor, float, optional):
                Time points at which to evaluate the Brier score of size n_times.
                Defaults to unique ``time``.
            weight (torch.Tensor, optional):
                Optional sample weight evaluated at ``time`` of size n_samples.
                Defaults to 1.
            weight_new_time (torch.tensor, optional):
                Optional sample weight evaluated at ``new_time`` of size n_times.
                Defaults to 1.

        Returns:
            torch.Tensor: Brier score evaluated at ``new_time``.

        Note:
            The function evaluates the time-dependent Brier score at time :math:`t \in \{t_1, \cdots, t_T\}` (argument ``new_time``).

            For each subject :math:`i \in \{1, \cdots, N\}`, denote :math:`X_i` as the survival time and :math:`D_i` as the
            censoring time. Survival data consist of the event indicator, :math:`\delta_i=(X_i\leq D_i)`
            (argument ``event``) and the time-to-event or censoring, :math:`T_i = \min(\{ X_i,D_i \})`
            (argument ``time``).

            The survival function, of subject :math:`i`
            is specified through :math:`S_i: [0, \infty) \rightarrow [0,1]`.
            The argument ``estimate`` is the estimated survival function. If ``new_time`` is specified, it should be of
            shape = (N,T) (:math:`(i,k)` th element is :math:`\hat{S}_i(t_k)`); if ``new_time`` is not specified,
            it should be of shape = (N,N) (:math:`(i,j)` th element is :math:`\hat{S}_i(T_j)`).

            The time-dependent Brier score :cite:p:`Graf1999` at time :math:`t` is the mean squared error of the event status

            .. math::

                BS(t) = \mathbb{E}\left[\left(1\left(X > t\right) - \hat{S}(t)\right)^2\right]

            The default Brier score estimate is

            .. math::

                \hat{BS}(t) = \frac{1}{n}\sum_i 1(T_i \leq t, \delta_i = 1) (0 - \hat{S}_i(t))^2 + 1(T_1 > t) (1- \hat{S}_i(t))^2

            To account for the fact that the event time are censored, the
            inverse probability weighting technique can be used. In this context,
            each subject associated with time
            :math:`t` is weighted by the inverse probability of censoring :math:`\omega(t) = 1 / \hat{D}(t)`, where
            :math:`\hat{D}(t)` is the Kaplan-Meier estimate of the censoring distribution, :math:`P(D>t)`.
            The censoring-adjusted Brier score is

            .. math::

                \hat{BS}(t) = \frac{1}{n}\sum_i \omega(T_i) 1(T_i \leq t, \delta_i = 1) (0 - \hat{S}_i(t))^2 + \omega(t) 1(T_1 > t) (1- \hat{S}_i(t))^2

            The censoring-adjusted Brier score can be obtained by specifying the argument
            ``weight``, the weights evaluated at each ``time`` (:math:`\omega(T_1), \cdots, \omega(T_N)`).
            If ``new_time`` is specified, the argument  ``weight_new_time``
            should also be specified accordingly, the weights evaluated at each ``new_time``
            (:math:`\omega(t_1), \cdots, \omega(t_K)`).
            In the context of train/test split, the weights should be derived from the censoring distribution estimated in the training data.


        Examples:
            >>> from torchsurv.stats.ipcw import get_ipcw
            >>> _ = torch.manual_seed(52)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.rand((n,len(time)))
            >>> brier_score = BrierScore()
            >>> brier_score(estimate, event, time)
            tensor([0.2463, 0.2740, 0.3899, 0.1964, 0.3608, 0.2821, 0.1932, 0.2978, 0.1950,
                    0.1668])
            >>> ipcw = get_ipcw(event, time) # ipcw at time
            >>> brier_score(estimate, event, time, weight=ipcw) # censoring-adjusted brier-score
            tensor([0.2463, 0.2740, 0.4282, 0.2163, 0.4465, 0.3826, 0.2630, 0.3888, 0.2219,
                    0.1882])
            >>> new_time = torch.unique(torch.randint(low=5, high=time.max().int(), size=(n*2,)).float())
            >>> ipcw_new_time = get_ipcw(event, time, new_time) # ipcw at new_time
            >>> estimate = torch.rand((n,len(new_time)))
            >>> brier_score(estimate, event, time, new_time, ipcw, ipcw_new_time) # censoring-adjusted brier-score at new time
            tensor([0.4036, 0.3014, 0.2517, 0.3947, 0.4200, 0.3908, 0.3766, 0.3737, 0.3596,
                    0.2088, 0.4922, 0.3237, 0.2255, 0.1841, 0.3029, 0.6919, 0.2357, 0.3507,
                    0.4364, 0.3312])

        References:

            .. bibliography::
                :filter: False

                Graf1999

        """

        # mandatory input format checks
        BrierScore._validate_brier_score_inputs(
            estimate, time, new_time, weight, weight_new_time
        )

        # update inputs as required
        (
            estimate,
            new_time,
            weight,
            weight_new_time,
        ) = BrierScore._update_brier_score_new_time(
            estimate, time, new_time, weight, weight_new_time
        )
        weight, weight_new_time = BrierScore._update_brier_score_weight(
            time, new_time, weight, weight_new_time
        )

        # further input format checks
        if self.checks:
            validate_inputs.validate_survival_data(event, time)
            validate_inputs.validate_evaluation_time(
                new_time, time, within_follow_up=False
            )
            validate_inputs.validate_estimate(estimate, time, new_time)

        # Calculating the residuals for each subject and time point
        residuals = torch.zeros_like(estimate)
        for index, new_time_i in enumerate(new_time):
            est = estimate[:, index]
            is_case = ((time <= new_time_i) & (event)).int()
            is_control = (time > new_time_i).int()

            residuals[:, index] = (
                torch.square(est) * is_case * weight
                + torch.square(1.0 - est) * is_control * weight_new_time[index]
            )

        # Calculating the brier scores at each time point
        brier_score = torch.mean(residuals, axis=0)

        # Create/overwrite internal attributes states
        if instate:
            # sort all objects by time
            self.order_time = torch.argsort(time, dim=0)
            self.time = time[self.order_time]
            self.event = event[self.order_time]
            self.weight = weight[self.order_time]
            self.new_time = new_time
            self.weight_new_time = weight_new_time
            self.estimate = torch.index_select(estimate, 0, self.order_time)
            self.brier_score = brier_score
            self.residuals = residuals

        return brier_score

    def integral(self):
        r"""Compute the integrated Brier Score.

        Returns:
            torch.Tensor: Integrated Brier Score.

        Examples:
            >>> _ = torch.manual_seed(52)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.rand((n,len(time)))
            >>> brier_score = BrierScore()
            >>> brier_score(estimate, event, time)
            tensor([0.2463, 0.2740, 0.3899, 0.1964, 0.3608, 0.2821, 0.1932, 0.2978, 0.1950,
                    0.1668])
            >>> brier_score.integral() # integrated brier score
            tensor(0.2862)

        Note:

            The integrated Brier score is the integral of the time-dependent Brier score over the interval
            :math:`[t_1, t_2]`, where :math:`t_1 = \min\left(\{T_i\}_{i = 1, \cdots, N}\right)` and :math:`t_2 = \max\left(\{T_i\}_{i = 1, \cdots, N}\right)`.
            It is defined by :cite:p:`Graf1999`

            .. math::

                    \hat{IBS} = \int_{t_1}^{t_2} \hat{BS}(t) dW(t)

            where :math:`W(t) = t / t_2`.

            The integral is estimated with the trapzoidal rule.

        """
        # Single time available
        if len(self.new_time) == 1:
            brier = self.brier_score[0]
        else:
            brier = torch.trapezoid(self.brier_score, self.new_time) / (
                self.new_time[-1] - self.new_time[0]
            )
        return brier

    def confidence_interval(
        self,
        method: str = "parametric",
        alpha: float = 0.05,
        alternative: str = "two_sided",
        n_bootstraps: int = 999,
    ) -> torch.Tensor:
        """Compute the confidence interval of the Brier Score.

        This function calculates either the pointwise confidence interval or the bootstrap
        confidence interval for the Brier Score. The pointwise confidence interval is computed
        assuming that the Brier score is normally distributed and using empirical standard errors.
        The bootstrap confidence interval is constructed based on the distribution of bootstrap samples.

        Args:
            method (str):
                Method for computing confidence interval. Defaults to "parametric".
                Must be one of the following: "parametric", "bootstrap".
            alpha (float):
                Significance level. Defaults to 0.05.
            alternative (str):
                Alternative hypothesis. Defaults to "two_sided".
                Must be one of the following: "two_sided", "greater", "less".
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor([lower,upper]):
                Lower and upper bounds of the confidence interval.

        Examples:
            >>> _ = torch.manual_seed(52)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.rand((n,len(time)))
            >>> brier_score = BrierScore()
            >>> brier_score(estimate, event, time)
            tensor([0.2463, 0.2740, 0.3899, 0.1964, 0.3608, 0.2821, 0.1932, 0.2978, 0.1950,
                    0.1668])
            >>> brier_score.confidence_interval() # default: parametric, two-sided
            tensor([[0.1061, 0.0604, 0.2360, 0.0533, 0.1252, 0.0795, 0.0000, 0.1512, 0.0381,
                     0.0051],
                    [0.3866, 0.4876, 0.5437, 0.3394, 0.5965, 0.4847, 0.4137, 0.4443, 0.3520,
                     0.3285]])
            >>> brier_score.confidence_interval(method = "bootstrap", alternative = "greater")
            tensor([[0.1455, 0.1155, 0.2741, 0.0903, 0.1985, 0.1323, 0.0245, 0.1938, 0.0788,
                     0.0440],
                    [1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
                     1.0000]])


        """

        assert (
            hasattr(self, "brier_score") and self.brier_score is not None
        ), "Error: Please calculate brier score using `BrierScore()` before calling `confidence_interval()`."

        if alternative not in ["less", "greater", "two_sided"]:
            raise ValueError(
                "'alternative' parameter must be one of ['less', 'greater', 'two_sided']."
            )

        if method == "bootstrap":
            conf_int = self._confidence_interval_bootstrap(
                alpha, alternative, n_bootstraps
            )
        elif method == "parametric":
            conf_int = self._confidence_interval_parametric(alpha, alternative)
        else:
            raise ValueError(
                f"Method {method} not implemented. Please choose either 'parametric' or 'bootstrap'."
            )
        return conf_int

    def p_value(
        self,
        method: str = "bootstrap",
        alternative: str = "two_sided",
        n_bootstraps: int = 999,
        null_value: float = None,
    ) -> torch.Tensor:
        """Perform a one-sample hypothesis test on the Brier score.

        This function calculates either the pointwise p-value or the bootstrap p-value
        for testing the null hypothesis that the estimated brier score is equal to
        bs0, where bs0 is the brier score that would be expected if the survival model
        was not providing accurate predictions beyond
        random chance. The pointwise p-value is computed assuming that the
        Brier score is normally distributed and using the empirical standard errors.
        To obtain the pointwise p-value, the Brier score under the null, bs0, must
        be provided.
        The bootstrap p-value is derived by permuting survival function's predictions
        to estimate the the sampling distribution under the null hypothesis.

        Args:
            method (str):
                Method for computing p-value. Defaults to "bootstrap".
                Must be one of the following: "parametric", "bootstrap".
            alternative (str):
                Alternative hypothesis. Defaults to "two_sided".
                Must be one of the following: "two_sided" (Brier score is not equal to bs0),
                "greater" (Brier score is greater than bs0), "less" (Brier score is less than bs0).
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ```method``` is not "bootstrap".
            null_value (float):
                The Brier score expected if the survival model was not
                providing accurate predictions beyond what would be beyond
                by random chance alone, i.e., bs0.


        Returns:
            torch.Tensor: p-value of the statistical test.

        Examples:
            >>> _ = torch.manual_seed(52)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> new_time = torch.unique(time)
            >>> estimate = torch.rand((n,len(new_time)))
            >>> brier_score = BrierScore()
            >>> brier_score(estimate, event, time, new_time)
            tensor([0.3465, 0.5310, 0.4222, 0.4582, 0.3601, 0.3395, 0.2285, 0.1975, 0.3120,
                    0.3883])
            >>> brier_score.p_value() # Default: bootstrap, two_sided
            tensor([1.0000, 0.0560, 1.0000, 1.0000, 1.0000, 1.0000, 0.8320, 0.8620, 1.0000,
                    1.0000])
            >>> brier_score.p_value(method = "parametric", alternative = "less", null_value = 0.3) # H0: bs = 0.3, Ha: bs < 0.3
            tensor([0.7130, 0.9964, 0.8658, 0.8935, 0.6900, 0.6630, 0.1277, 0.1128, 0.5383,
                    0.8041])

        """

        assert (
            hasattr(self, "brier_score") and self.brier_score is not None
        ), "Error: Please calculate the brier score using `BrierScore()` before calling `p_value()`."

        if alternative not in ["less", "greater", "two_sided"]:
            raise ValueError(
                "'alternative' parameter must be one of ['less', 'greater', 'two_sided']."
            )

        if method == "parametric" and null_value is None:
            raise ValueError(
                "Error: If the method is 'parametric', you must provide the 'null_value'."
            )

        if method == "parametric":
            pvalue = self._p_value_parametric(alternative, null_value)
        elif method == "bootstrap":
            pvalue = self._p_value_bootstrap(alternative, n_bootstraps)
        else:
            raise ValueError(
                f"Method {method} not implemented. Please choose either 'parametric' or 'bootstrap'."
            )
        return pvalue

    def compare(
        self, other, method: str = "parametric", n_bootstraps: int = 999
    ) -> torch.tensor:
        """Compare two Brier scores.

        This function compares two Brier scores computed on the
        same data with different risk scores. The statistical hypotheses are
        formulated as follows, null hypothesis: brierscore1 = brierscore2 and alternative
        hypothesis: brierscore1 < brierscore2.
        The statistical test is either a Student t-test for paired samples or a two-sample bootstrap test.
        The Student t-test for paired samples assumes that the Brier Scores are normally distributed
        and uses the Brier scores' empirical standard errors.

        Args:
            other (BrierScore):
                Another instance of the BrierScore class representing brierscore2.
            method (str):
                Statistical test used to perform the hypothesis test. Defaults to "parametric".
                Must be one of the following: "parametric", "bootstrap".
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.tensor: p-value of the statistical test.

        Examples:
            >>> _ = torch.manual_seed(52)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> brier_score = BrierScore()
            >>> brier_score(torch.rand((n,len(time))), event, time)
            tensor([0.2463, 0.2740, 0.3899, 0.1964, 0.3608, 0.2821, 0.1932, 0.2978, 0.1950,
                    0.1668])
            >>> brier_score2 = BrierScore()
            >>> brier_score2(torch.rand((n,len(time))), event, time)
            tensor([0.4136, 0.2750, 0.3002, 0.2826, 0.2030, 0.2643, 0.2525, 0.2964, 0.1804,
                    0.3109])
            >>> brier_score.compare(brier_score2) # default: parametric
            tensor([0.1793, 0.4972, 0.7105, 0.1985, 0.9254, 0.5591, 0.3455, 0.5060, 0.5437,
                    0.0674])
            >>> brier_score.compare(brier_score2, method = "bootstrap")
            tensor([0.1360, 0.5030, 0.7310, 0.2090, 0.8630, 0.5490, 0.3120, 0.5110, 0.5460,
                    0.1030])

        """

        assert (
            hasattr(self, "brier_score") and self.brier_score is not None
        ), "Error: Please calculate the brier score using `BrierScore()` before calling `compare()`."

        # assert that the same data were used to compute the two brier score
        if torch.any(self.event != other.event) or torch.any(self.time != other.time):
            raise ValueError(
                "Mismatched survival data: 'time' and 'event' should be the same for both brier score computations."
            )
        if torch.any(self.new_time != other.new_time):
            raise ValueError(
                "Mismatched evaluation times: 'new_time' should be the same for both brier score computations."
            )

        if method == "parametric":
            pvalue = self._compare_parametric(other)
        elif method == "bootstrap":
            pvalue = self._compare_bootstrap(other, n_bootstraps)
        else:
            raise ValueError(
                "Method not implemented. Please choose either 'parametric' or 'bootstrap'."
            )
        return pvalue

    def _brier_score_se(self):
        """Brier Score's empirical standard errors."""

        return torch.std(self.residuals, axis=0) / (self.time.shape[0] ** (1 / 2))

    def _confidence_interval_parametric(
        self, alpha: float, alternative: str
    ) -> torch.Tensor:
        """Confidence interval of Brier score assuming that the Brier score
        is normally distributed and using empirical standard errors.
        """

        alpha = alpha / 2 if alternative == "two_sided" else alpha

        brier_score_se = self._brier_score_se()

        if torch.all(brier_score_se) > 0:
            ci = (
                -torch.distributions.normal.Normal(0, 1).icdf(torch.tensor(alpha))
                * brier_score_se
            )
            lower = torch.max(torch.tensor(0.0), self.brier_score - ci)
            upper = torch.min(torch.tensor(1.0), self.brier_score + ci)

            if alternative == "less":
                lower = torch.zeros_like(lower)
            elif alternative == "greater":
                upper = torch.ones_like(upper)
        else:
            raise ValueError(
                "The standard error of the brier score must be a positive value."
            )

        return torch.stack([lower, upper], dim=0)

    def _confidence_interval_bootstrap(
        self, alpha: float, alternative: str, n_bootstraps: int
    ) -> torch.tensor:
        """Bootstrap confidence interval of the Brier Score using Efron percentile method.

        References:
            Efron, Bradley; Tibshirani, Robert J. (1993).
                An introduction to the bootstrap, New York: Chapman & Hall, software.
        """

        # brier score given bootstrap distribution
        brier_score_bootstrap = self._bootstrap_brier_score(
            metric="confidence_interval", n_bootstraps=n_bootstraps
        )

        # intialize tensor to store confidence intervals
        lower = torch.zeros_like(self.brier_score)
        upper = torch.zeros_like(self.brier_score)

        # iterate over time
        for index_t in range(len(self.brier_score)):
            # obtain confidence interval
            if alternative == "two_sided":
                lower[index_t], upper[index_t] = torch.quantile(
                    brier_score_bootstrap[:, index_t],
                    torch.tensor(
                        [alpha / 2, 1 - alpha / 2], device=self.brier_score.device
                    ),
                )
            elif alternative == "less":
                upper[index_t] = torch.quantile(
                    brier_score_bootstrap[:, index_t],
                    torch.tensor(1 - alpha, device=self.brier_score.device),
                )
                lower[index_t] = torch.tensor(0.0, device=self.brier_score.device)
            elif alternative == "greater":
                lower[index_t] = torch.quantile(
                    brier_score_bootstrap[:, index_t],
                    torch.tensor(alpha, device=self.brier_score.device),
                )
                upper[index_t] = torch.tensor(1.0, device=self.brier_score.device)

        return torch.stack([lower, upper], dim=0)

    def _p_value_parametric(
        self, alternative: str, null_value: float = 0.5
    ) -> torch.tensor:
        """p-value for a one-sample hypothesis test of the Brier score
        assuming that the Brier score is normally distributed and using empirical standard error.
        """

        brier_score_se = self._brier_score_se()

        # get p-value
        if torch.all(brier_score_se) > 0:
            p = torch.distributions.normal.Normal(0, 1).cdf(
                (self.brier_score - null_value) / brier_score_se
            )
            if alternative == "two_sided":
                mask = self.brier_score >= 0.5
                p[mask] = 1 - p[mask]
                p *= 2
                p = torch.min(
                    torch.tensor(1.0, device=self.brier_score.device), p
                )  # in case critical value is below 0.5
            elif alternative == "greater":
                p = 1 - p
        else:
            raise ValueError(
                "The standard error of the brier score must be a positive value."
            )

        return p

    def _p_value_bootstrap(self, alternative, n_bootstraps) -> torch.tensor:
        """p-value for a one-sample hypothesis test of the Brier score using
        permutation of survival distribution prediction to estimate sampling distribution under the null
        hypothesis.
        """

        # brier score bootstraps given null distribution
        brierscore0 = self._bootstrap_brier_score(
            metric="p_value", n_bootstraps=n_bootstraps
        )

        # intialize empty tensor to store p-values
        p_values = torch.zeros_like(self.brier_score)

        # iterate over time
        for index_t, brier_score_t in enumerate(self.brier_score):
            # Derive p-value
            p = (1 + torch.sum(brierscore0[:, index_t] <= brier_score_t)) / (
                n_bootstraps + 1
            )
            if alternative == "two_sided":
                if brier_score_t >= 0.5:
                    p = 1 - p
                p *= 2
                p = torch.min(
                    torch.tensor(1.0, device=self.brier_score.device), p
                )  # in case very small bootstrap sample size is used
            elif alternative == "greater":
                p = 1 - p

            p_values[index_t] = p

        return p_values

    def _compare_parametric(self, other):
        """Student t-test for paired samples assuming that
        the Brier scores are normally distributed and using
        empirical standard errors."""

        # sample size
        n_samples = self.time.shape[0]

        # intialize empty vector to store p_values
        p_values = torch.zeros_like(self.brier_score)

        # iterate over time
        for index_t, brier_score_t in enumerate(self.brier_score):
            # compute standard error of the difference
            paired_se = torch.std(
                self.residuals[:, index_t] - other.residuals[:, index_t]
            ) / (n_samples ** (1 / 2))

            # compute t-stat
            t_stat = (brier_score_t - other.brier_score[index_t]) / paired_se

            # p-value
            p_values[index_t] = torch.tensor(
                stats.t.cdf(
                    t_stat, df=n_samples - 1
                ),  # student-t cdf not available on torch
                dtype=self.brier_score.dtype,
                device=self.brier_score.device,
            )

        return p_values

    def _compare_bootstrap(self, other, n_bootstraps):
        """Boostrap two-sample test to compare two Brier scores."""

        # bootstrap brier scores given null hypothesis that brierscore1 and
        # brierscore2 come from the same distribution
        brier_score1_null = self._bootstrap_brier_score(
            metric="compare", other=other, n_bootstraps=n_bootstraps
        )
        brier_score2_null = self._bootstrap_brier_score(
            metric="compare", other=other, n_bootstraps=n_bootstraps
        )

        # bootsrapped test statistics
        t_boot = brier_score1_null - brier_score2_null

        # observed test statistics
        t_obs = self.brier_score - other.brier_score

        # intialize empty tensor to store p-values
        p_values = torch.zeros_like(self.brier_score)

        # iterate over time
        for index_t, _ in enumerate(self.brier_score):
            p_values[index_t] = (
                1 + torch.sum(t_boot[:, index_t] <= t_obs[index_t])
            ) / (n_bootstraps + 1)

        return p_values

    def _bootstrap_brier_score(
        self, metric: str, n_bootstraps: int, other=None
    ) -> torch.tensor:
        """Compute bootstrap samples of the Brier Score.

        Args:
            metric (str): Must be one of the following: "confidence_interval", "compare".
                If "confidence_interval", computes bootstrap
                samples of the Brier score given the data distribution. If "compare", computes
                bootstrap samples of the Brier score given the sampling distribution under the comparison test
                null hypothesis (brierscore1 = brierscore2).
            n_bootstraps (int): Number of boostrap samples.
            other (optional, BrierScore):
                Another instance of the BrierScore class representing brierscore2.
                Only required if ``metric`` is "compare".


        Returns:
            torch.tensor: Bootstrap samples of Brier score.
        """

        # Initiate empty list to store brier score
        brier_scores = []

        # Get the bootstrap samples of brier score
        for _ in range(n_bootstraps):
            if (
                metric == "confidence_interval"
            ):  # bootstrap samples given data distribution
                index = torch.randint(
                    low=0,
                    high=self.estimate.shape[0],
                    size=(self.estimate.shape[0],),
                )
                brier_scores.append(
                    self(
                        self.estimate[index, :],
                        self.event[index],
                        self.time[index],
                        self.new_time,
                        self.weight[index],
                        self.weight_new_time,
                        instate=False,
                    )
                )  # Run without saving internal state
            elif (
                metric == "compare"
            ):  # bootstrap samples given null distribution (brierscore1 = brierscore2)
                index = torch.randint(
                    low=0,
                    high=self.estimate.shape[0] * 2,
                    size=(self.estimate.shape[0],),
                )

                # with prob 0.5, take the weight_new_time from self and with prob 0.5 from other
                weight_new_time = (
                    self.weight_new_time
                    if torch.rand(1) < 0.5
                    else other.weight_new_time
                )

                brier_scores.append(
                    self(  # sample with replacement from pooled sample
                        torch.cat((self.estimate, other.estimate))[index, :],
                        torch.cat((self.event, other.event))[index],
                        torch.cat((self.time, other.time))[index],
                        self.new_time,
                        torch.cat((self.weight, other.weight))[index],
                        weight_new_time,
                        instate=False,
                    )
                )
            elif (
                metric == "p_value"
            ):  # bootstrap samples given null distribution (estimate are not informative)
                estimate = copy.deepcopy(self.estimate)
                estimate = estimate[
                    torch.randperm(estimate.shape[0]), :
                ]  # Shuffle estimate
                brier_scores.append(
                    self(
                        estimate,
                        self.event,
                        self.time,
                        self.new_time,
                        self.weight,
                        self.weight_new_time,
                        instate=False,
                    )
                )  # Run without saving internal state

        brier_scores = torch.stack(brier_scores, dim=0)

        if torch.any(torch.isnan(brier_scores)):
            raise ValueError(
                "The brier score computed using bootstrap should not be NaN."
            )

        return brier_scores

    @staticmethod
    def _find_torch_unique_indices(
        inverse_indices: torch.tensor, counts: torch.tensor
    ) -> torch.tensor:
        """return unique_sorted_indices such that
        sorted_unique_tensor[inverse_indices] = original_tensor
        original_tensor[unique_sorted_indices] = sorted_unique_tensor

        Usage:
            _, inverse_indices, counts = torch.unique(
                    x, sorted=True, return_inverse=True, return_counts=True
            )
            sorted_unique_indices = Auc._find_torch_unique_indices(
                    inverse_indices, counts
            )
        """

        _, ind_sorted = torch.sort(inverse_indices, stable=True)
        cum_sum = counts.cumsum(0)
        cum_sum = torch.cat((torch.tensor([0]), cum_sum[:-1]))
        return ind_sorted[cum_sum]

    @staticmethod
    def _validate_brier_score_inputs(
        estimate: torch.tensor,
        time: torch.tensor,
        new_time: torch.tensor,
        weight: torch.tensor,
        weight_new_time: torch.tensor,
    ) -> torch.tensor:
        # check new_time and weight are provided, weight_new_time should be provided
        if all([new_time is not None, weight is not None, weight_new_time is None]):
            raise ValueError(
                "Please provide 'weight_new_time', the weight evaluated at 'new_time'."
            )

        # check that estimate has 2 dimensions estimate are probabilities
        if torch.any(estimate < 0) or torch.any(estimate > 1):
            raise ValueError(
                "The 'estimate' input should contain estimated survival probabilities between 0 and 1."
            )

        # check if estimate is of the correct dimension
        if estimate.ndim != 2:
            raise ValueError("The 'estimate' input should have two dimensions.")

        # check if new_time are not specified and estimate are not evaluated at time
        if new_time is None and len(time) != estimate.shape[1]:
            raise ValueError(
                "Mismatched dimensions: The number of columns in 'estimate' does not match the length of 'time'. "
                "Please provide the times at which 'estimate' is evaluated using the 'new_time' input."
            )

    @staticmethod
    def _update_brier_score_new_time(
        estimate: torch.tensor,
        time: torch.tensor,
        new_time: torch.tensor,
        weight: torch.tensor,
        weight_new_time: torch.tensor,
    ) -> torch.tensor:
        # check format of new_time
        if (
            new_time is not None
        ):  # if new_time are specified: ensure it has the correct format
            if isinstance(new_time, int):
                new_time = torch.tensor([new_time]).float()

            if new_time.ndim == 0:
                new_time = new_time.unsqueeze(0)

        else:  # else: find new_time
            # if new_time are not specified, use unique time
            new_time, inverse_indices, counts = torch.unique(
                time, sorted=True, return_inverse=True, return_counts=True
            )
            sorted_unique_indices = BrierScore._find_torch_unique_indices(
                inverse_indices, counts
            )

            # for time-dependent estimate, select those corresponding to new time
            estimate = estimate[:, sorted_unique_indices]

            if weight is not None:
                # select weight corresponding at new time
                weight_new_time = weight[sorted_unique_indices]

        return estimate, new_time, weight, weight_new_time

    @staticmethod
    def _update_brier_score_weight(
        time: torch.tensor,
        new_time: torch.tensor,
        weight: torch.tensor,
        weight_new_time: torch.tensor,
    ) -> torch.tensor:
        # if weight was not specified, weight of 1
        if weight is None:
            weight = torch.ones_like(time)
            weight_new_time = torch.ones_like(new_time)

        return weight, weight_new_time


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



import collections
import copy
import sys
from typing import Callable

import torch
from torch import nn


class Momentum(nn.Module):
    r"""
    Survival framework to momentum update learning to decouple batch size during model training.
    Two networks are concurently trained, an online network and a target network. The online network outputs batches
    are concanetaed and used by the target network, so it virtually increase its batchsize.

    The target network (k)is updated using an exponential momentum average (**EMA**) using parameters from the online network (q).
    The online network is trained using a memory bank of previously computed log hazards, but only tracking loss from current batch.

    .. math::
            \theta_k \leftarrow m \theta_k + (1 - m) \theta_q

    with m=0.999.


    Here is a pseudo python code to illustrate what is going on under the hood.

    .. highlight:: python
    .. code-block:: python

        model_q = model_k                 # Same architecture, random weights
        model_k.require_grad = False      # No gradient update for target network (k)
        hz_memory_bank = deque(maxlen=n * batch_size)  # Double-ended queue size n * batch_size

        for epoch in epochs:
            hz_q = model_q(batch)            # Compute current estmate w/ ONLINE network (q)
            hz_loss = hz_memory_bank + hz_q  # Combine current log hz and memory bank
            loss = loss_function(hz_loss)    # Compute loss with pooled log hz
            loss.backward()                  # Update online model (q) w/ PyTorch autograd
            model_k.ema_update(model_q)      # Update target model (k) with Exponential Moving Average (EMA)
            hz_k = model_k(batch)            # Compute batch estimate w/ TARGET network (k)
            hz_memory_bank += hz_k           # Replace oldest batch with current from memory bank

    Note:
        This code is inspired from MoCo :cite:p:`he2019` and its ability to decouple batch size from training size.

    Note:
        A notable difference is that memory bank is updated at the end of the step, not at the beginning. This is because
        MoCo uses the target network batch for the positive pair, while we use a rank-based loss function. We then need
        to exclude the current batch from the memory bank to effectively compute our loss.


    References:
        .. bibliography::
            :filter: False

            he2019

    """

    def __init__(
        self,
        backbone: nn.Module,
        loss: Callable,
        batchsize: int = 16,
        steps: int = 4,
        rate: float = 0.999,
    ):
        """
        Initialise the momentum class. Use must provide their model as backbone.

        Args:
            backbone (nn.Module):
                Torch model to be use as backbone. The model must return either one (Cox) or two ouputs (Weibull)
            loss (Callable): Torchsurv loss function (Cox, Weibull)
            batchsize (int, optional):
                Number of samples per batch. Defaults to 16.
            n (int, optional):
                Number of queued batches to be stored for training. Defaults to 4.
            rate (float, optional):
                Exponential moving average rate. Defaults to 0.999.

        Examples:
            >>> from torchsurv.loss import cox, weibull
            >>> _ = torch.manual_seed(42)
            >>> n = 4
            >>> params = torch.randn((n, 16))
            >>> events = torch.randint(low=0, high=2, size=(n,), dtype=torch.bool)
            >>> times = torch.randint(low=1, high=100, size=(n,))
            >>> backbone = torch.nn.Sequential(torch.nn.Linear(16, 1))  # Cox expect one ouput
            >>> model = Momentum(backbone=backbone, loss=cox.neg_partial_log_likelihood)
            >>> model(params, events, times)
            tensor(0.0978, grad_fn=<DivBackward0>)
            >>> model.online(params)  # online network (q) - w/ gradient
            tensor([[-0.7867],
                    [ 0.3161],
                    [-1.2158],
                    [-0.8195]], grad_fn=<AddmmBackward0>)
            >>> model.target(params)  # target network (k) - w/o gradient
            tensor([[-0.7867],
                    [ 0.3161],
                    [-1.2158],
                    [-0.8195]])

        Note:
            `self.encoder_k` is the recommended to be used for inference. It refers to the target network (momentum).
        """
        super().__init__()
        self.rate = rate
        self.online = copy.deepcopy(backbone)  # q >> online network
        self.target = copy.deepcopy(backbone)  # k >> target network (momentum)
        self._init_encoder_k()
        self.loss = loss

        # survival data structure
        self.survtuple = collections.namedtuple(
            "survival", ["estimate", "event", "time"]
        )

        # Hazards: current batch & memory
        self.memory_q = collections.deque(maxlen=batchsize)  # online (q)
        self.memory_k = collections.deque(maxlen=batchsize * steps)  # target (k)

    def forward(
        self, inputs: torch.Tensor, event: torch.Tensor, time: torch.Tensor
    ) -> torch.Tensor:
        """Compute the loss for the current batch and update the memory bank using momentum class.

        Args:
            inputs (torch.Tensor): Input tensors to the backbone model
            event (torch.Tensor): A boolean tensor indicating whether a patient experienced an event.
            time (torch.Tensor): A positive float tensor representing time to event (or censoring time)

        Returns:
            torch.Tensor: A loss tensor for the current batch.

        Examples:
            >>> from torchsurv.loss import cox, weibull
            >>> _ = torch.manual_seed(42)
            >>> n = 128  # samples
            >>> x = torch.randn((n, 16))
            >>> y = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> t = torch.randint(low=1, high=100, size=(n,))
            >>> backbone = torch.nn.Sequential(torch.nn.Linear(16, 1))  # (log hazards)
            >>> model_cox = Momentum(backbone, loss=cox.neg_partial_log_likelihood)  # Cox loss
            >>> with torch.no_grad(): model_cox.forward(x, y, t)
            tensor(2.1366)
            >>> backbone = torch.nn.Sequential(torch.nn.Linear(16, 2))  # (lambda, rho)
            >>> model_weibull = Momentum(backbone, loss=weibull.neg_log_likelihood)  # Weibull loss
            >>> with torch.no_grad(): torch.round(model_weibull.forward(x, y, t), decimals=2)
            tensor(68.0400)

        """

        estimate_q = self.online(inputs)
        for estimate in zip(estimate_q, event, time):
            self.memory_q.append(self.survtuple(*list(estimate)))
        loss = self._bank_loss()
        with torch.no_grad():
            self._update_momentum_encoder()
            estimate_k = self.target(inputs)
            for estimate in zip(estimate_k, event, time):
                self.memory_k.append(self.survtuple(*list(estimate)))
        return loss

    @torch.no_grad()  # deactivates autograd
    def infer(self, inputs: torch.Tensor) -> torch.Tensor:
        """Evaluate data with target network

        Args:
            x (torch.Tensor): Input tensors to the backbone model

        Returns:
            torch.Tensor: Predictions from target (momentum) network without augmentation (.eval()) nor gradient.

        Examples:
            >>> from torchsurv.loss import weibull
            >>> _ = torch.manual_seed(42)
            >>> backbone = torch.nn.Sequential(torch.nn.Linear(8, 2))  # Weibull expect two ouputs
            >>> model = Momentum(backbone=backbone, loss=weibull.neg_log_likelihood)
            >>> model.infer(torch.randn((3, 8)))
            tensor([[ 0.5342,  0.0062],
                    [ 0.6439,  0.7863],
                    [ 0.9771, -0.8513]])

        """
        self.target.eval()  # notify all your layers that you are in eval mode
        return self.target(inputs)

    def _bank_loss(self) -> torch.Tensor:
        """computer the  negative loss likelyhood from memory bank"""

        # Combine current batch and momentum
        bank = self.memory_k + self.memory_q
        assert all(
            x in bank[0]._fields for x in ["estimate", "event", "time"]
        ), "All fields must be present"
        return self.loss(
            torch.stack([mem.estimate.cpu() for mem in bank]).squeeze(),
            torch.stack([mem.event.cpu() for mem in bank]).squeeze(),
            torch.stack([mem.time.cpu() for mem in bank]).squeeze(),
        )

    @torch.no_grad()
    def _update_momentum_encoder(self):
        """Exponantial moving average"""
        for param_b, param_m in zip(self.online.parameters(), self.target.parameters()):
            param_m.data = param_m.data * self.rate + param_b.data * (1.0 - self.rate)

    @torch.no_grad()
    def _init_encoder_k(self):
        for param_q, param_k in zip(self.online.parameters(), self.target.parameters()):
            param_k.data.copy_(param_q.data)
            param_k.requires_grad = False


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



import unittest

# Global
import lightning as L
import torch
from loguru import logger
from utils import (
    LitSurvival,
    LitSurvivalTwins,
    SimpleLinearNNOneParameter,
    SimpleLinearNNTwoParameters,
)

from torchsurv.loss.weibull import neg_log_likelihood as weibull

# Set a seed for reproducibility
seed_value = 45
torch.manual_seed(seed_value)
BATCH_N = 10


trainer = L.Trainer(max_epochs=2, log_every_n_steps=5)


class TestLitTraining(unittest.TestCase):
    def run_training(self, trainer, model):
        logger.critical(f"Loading {model.dataname} dataset")
        trainer.fit(model)
        return model

    def test_one_param(self):
        """One parameter Weibull"""
        model = self.run_training(
            trainer,
            LitSurvival(
                backbone=SimpleLinearNNOneParameter(input_size=2),
                loss=weibull,
                dataname="lung",
                batch_size=248 // BATCH_N,
            ),
        )
        with torch.no_grad():
            params = model(torch.randn(1, 2))
        self.assertEqual(params.size(), (1, 1))

    def test_two_params(self):
        """Two parameters Weibull"""
        model = self.run_training(
            trainer,
            LitSurvival(
                backbone=SimpleLinearNNTwoParameters(input_size=2),
                loss=weibull,
                dataname="gbsb",
                batch_size=686 // BATCH_N,
            ),
        )
        with torch.no_grad():
            params = model(torch.randn(1, 2))
        self.assertEqual(params.size(), (1, 2))

    def test_twins(self):
        """Two parameters Weibull"""
        model = LitSurvivalTwins(
            backbone=SimpleLinearNNOneParameter(input_size=2),
            loss=weibull,
            steps=2,
            dataname="gbsb",
            batch_size=686 // BATCH_N,
        )
        res = self.run_training(
            trainer,
            model,
        )

        with torch.no_grad():
            x = torch.randn(1, 2)
            # logger.info(f"No training: {res.twinnets.backbone(x)}")
            # logger.info(f"Student (Q): {res.twinnets.encoder_q(x)}")
            # logger.info(f"Teacher (K): {res.twinnets.encoder_k(x)}")

        # Model parameters (sanity checks)
        # for params in model.twinnets.backbone.parameters():
        #     logger.info(f"No training: {params}")
        # for params in model.twinnets.encoder_q.parameters():
        #     logger.info(f"Student (Q): {params}")
        # for params in model.twinnets.encoder_k.parameters():
        #     logger.info(f"Teacher (K): {params}")


if __name__ == "__main__":
    unittest.main()



import unittest

import torch
from pytorch_lightning import LightningModule, Trainer
from torch import nn
from torch.utils.data import DataLoader, random_split
from torchvision.datasets import MNIST
from torchvision.transforms import v2

from torchsurv.loss.cox import neg_partial_log_likelihood
from torchsurv.metrics.cindex import ConcordanceIndex


class LitMNIST(LightningModule):
    def __init__(self, hidden_size=128, learning_rate=5e-4):
        super().__init__()

        # Set our init args as class attributes
        self.data_dir = "."
        self.hidden_size = hidden_size
        self.learning_rate = learning_rate
        self.batch_size = 256
        self.num_workers = 4

        # Model optimizisation
        self.loss = neg_partial_log_likelihood
        self.cindex = ConcordanceIndex()

        # Hardcode some dataset specific attributes
        self.dims = (1, 28, 28)
        channels, width, height = self.dims
        self.transform = v2.Compose(
            [
                v2.ToImage(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=(0,), std=(1,)),
                v2.ColorJitter(brightness=0.2, contrast=0.2, saturation=0, hue=0),
            ]
        )

        # Define PyTorch model
        self.model = nn.Sequential(
            nn.Flatten(),
            nn.Linear(channels * width * height, hidden_size),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, 1),
        )

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        y[y == 0] = 10.0  # Offset 0 to prevent log(0)
        params = self(x)
        loss = self.loss(params, torch.ones_like(y, device=y.device).bool(), y)
        self.log("loss", loss, prog_bar=True, on_step=True, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y[y == 0] = 10  # Offset 0 to prevent log(0)
        params = self(x)
        loss = self.loss(params, torch.ones_like(y, device=y.device).bool(), y)
        cindex = self.cindex(
            params, torch.ones_like(y, device=y.device).bool(), y.float()
        )
        self.log("val_loss", loss, prog_bar=True, on_step=True, on_epoch=True)
        self.log(
            "cindex",
            cindex,
            prog_bar=True,
            on_step=True,
            on_epoch=True,
        )
        return loss

    def test_step(self, batch, batch_idx):
        # Here we just reuse the validation_step for testing
        return self.validation_step(batch, batch_idx)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer

    ####################
    # DATA RELATED HOOKS
    ####################

    def prepare_data(self):
        # download
        MNIST(self.data_dir, train=True, download=True)
        MNIST(self.data_dir, train=False, download=True)

    def setup(self, stage=None):
        # Assign train/val datasets for use in dataloaders
        if stage == "fit" or stage is None:
            mnist_full = MNIST(self.data_dir, train=True, transform=self.transform)
            self.mnist_train, self.mnist_val = random_split(mnist_full, [55000, 5000])

        # Assign test dataset for use in dataloader(s)
        if stage == "test" or stage is None:
            self.mnist_test = MNIST(
                self.data_dir, train=False, transform=self.transform
            )

    def train_dataloader(self):
        return DataLoader(
            self.mnist_train,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=True,
        )

    def val_dataloader(self):
        return DataLoader(
            self.mnist_val,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=True,
        )

    def test_dataloader(self):
        return DataLoader(
            self.mnist_test,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=True,
        )


class TestMetrics(unittest.TestCase):
    def test_training(self):
        model = LitMNIST()
        trainer = Trainer(
            logger=False,
            enable_checkpointing=False,
            accelerator="auto",
            fast_dev_run=2,
        )
        trainer.fit(model)
        results = trainer.validate(model)[0]
        values = list(results.values())
        names = list(results.keys())
        self.assertTrue(names == ["val_loss_epoch", "cindex_epoch"])
        self.assertTrue(values[0] > 0)  # Loss
        self.assertTrue(all([values[1] <= 1, values[1] >= 0]))  # Cindex


if __name__ == "__main__":
    unittest.main()



import sys

import torch

TORCH_CLAMP_VALUE = 1e10


def neg_log_likelihood(
    log_params: torch.Tensor,
    event: torch.Tensor,
    time: torch.Tensor,
    reduction: str = "mean",
    checks: bool = True,
) -> torch.Tensor:
    r"""
    Negative of the log likelihood for the Weibull Accelerated Time Failure (AFT) survival model.

    Args:
        log_params (torch.Tensor, float):
            Parameters of the Weibull distribution of shape = (n_samples, 1) or (n_samples, 2).
            The first column corresponds to the log scale parameter. The second column
            corresponds to the log shape parameter.  If the log shape parameter is missing, it is
            imputed with 0.
        event (torch.Tensor, bool):
            Event indicator of length n_samples (= True if event occured).
        time (torch.Tensor, float):
            Time-to-event or censoring of length n_samples.
        reduction (str):
            Method to reduce losses. Defaults to "mean".
            Must be one of the following: "sum", "mean".
        checks (bool):
            Whether to perform input format checks.
            Enabling checks can help catch potential issues in the input data.
            Defaults to True.

    Returns:
        (torch.Tensor, float): Negative of the log likelihood.

    Note:

        For each subject :math:`i \in \{1, \cdots, N\}`, denote :math:`X_i` as the survival time and :math:`D_i` as the
        censoring time. Survival data consist of the event indicator, :math:`\delta_i=1(X_i\leq D_i)`
        (argument ``event``) and the time-to-event or censoring, :math:`T_i = \min(\{ X_i,D_i \})`
        (argument ``time``).

        The log hazard function for the Weibull AFT survival model :cite:p:`Carroll2003` of subject :math:`i` at time :math:`t` has the form:

        .. math::

            \log h_i(t) = \log{\rho_i} - \log{\lambda_i} + (\rho_i -1) \left( \log{t} - \log{\lambda_i}\right)

        where :math:`\log{\lambda_i}` is the log scale parameter (first column of argument ``log_params``)
        and :math:`\log{\rho_i}` is the log shape parameter (second column of argument ``log_params``).
        The cumulative hazard for the Weibull survival model of subject :math:`i` at time :math:`t` has the form:

        .. math::

            H_i(t) = \left(\frac{t}{\lambda_i}\right)^{\rho_i}

        The survival function for the Weibull survival model of subject :math:`i` at time :math:`t` has the form:

        .. math::

            S_i(t) = 1 - F(t | \lambda_i, \rho_i)

        where :math:`F(t | \lambda, \rho)` is the cumulative distribution function (CDF) of the Weibull distribution given
        scale parameter :math:`\lambda` and shape parameter :math:`\rho`.

        The log likelihood of the Weibull survival model is

        .. math::

            ll = \sum_{i: \delta_i = 1} \log h_i(T_i) - \sum_{i = 1}^N H_i(T_i)

    Examples:
        >>> _ = torch.manual_seed(42)
        >>> n = 4
        >>> log_params = torch.randn((n, 2))
        >>> event = torch.randint(low=0, high=2, size=(n,), dtype=torch.bool)
        >>> time = torch.randint(low=1, high=100, size=(n,))
        >>> neg_log_likelihood(log_params, event, time) # Default: mean of log likelihoods across subject
        tensor(47.5035)
        >>> neg_log_likelihood(log_params, event, time, reduction = 'sum') # Sum of log likelihoods across subject
        tensor(190.0141)
        >>> neg_log_likelihood(torch.randn((n, 1)), event, time)  # Missing shape: exponential decrease
        tensor(66.7203)

    References:

        .. bibliography::
            :filter: False

            Carroll2003

    """

    if checks:
        _check_inputs(log_params, event, time)

    # Negative log likelihood
    nll = torch.neg(
        event * log_hazard(log_params, time, all_times=False)
        - cumulative_hazard(log_params, time, all_times=False)  # Huge values here
    )

    if any(torch.isinf(nll)):
        # Remove any torch.inf values
        nll = nll[~torch.isinf(nll)]

    if reduction.lower() == "mean":
        loss = nll.nanmean()
    elif reduction.lower() == "sum":
        loss = nll.sum()
    else:
        raise (
            ValueError(
                f"Reduction {reduction} is not implemented yet, should be one of ['mean', 'sum']."
            )
        )
    return loss


def survival_function(
    log_params: torch.Tensor, time: torch.Tensor, all_times: bool = True
) -> torch.Tensor:
    """Survival function for the Weibull Accelerated Time Failure (AFT) survival model.

    Args:
        log_params (torch.Tensor, float):
            Parameters of the Weibull distribution of shape = (n_samples, 1) or (n_samples, 2).
            The first column corresponds to the log scale parameter. The second column
            corresponds to the log shape parameter. If the log shape parameter is missing, it is
            imputed with 0.
        time (torch.Tensor, float):
            Time at which to evaluate the survival function.
            Should be of length n_samples to evaluate the survival function at observed time-to-event or censoring,
            or of length one to evaluate the survival function at a new time.
        all_times (bool):
            If True, subject-specific survival function is evaluated at all ``time`` (used for evaluation metrics).
            If False, subject-specific survival function is evaluated at respective ``time``.
            Defaults is True.
            Ignored if ``time`` is of length one.

    Returns:
        (torch.Tensor, float): Subject-specific survival function evaluated at ``time``.

    Examples:
        >>> _ = torch.manual_seed(42)
        >>> time = torch.randint(low=1, high=100, size=(4,))
        >>> log_params = torch.randn((4, 2))
        >>> survival_function(log_params, time, all_times = False)  # Survival at respective time
        tensor([0.0002, 0.0000, 0.0299, 0.0000])
        >>> survival_function(log_params, time, all_times = True)  # Default. Survival at all observed time
        tensor([[1.7941e-04, 0.0000e+00, 0.0000e+00, 0.0000e+00],
                [2.8610e-06, 0.0000e+00, 0.0000e+00, 0.0000e+00],
                [4.1870e-01, 3.1040e-02, 2.9881e-02, 6.8224e-02],
                [9.5576e-04, 0.0000e+00, 0.0000e+00, 0.0000e+00]])
        >>> survival_function(log_params, time=torch.tensor(10.0))  # Survival at one new time (e.g., 10 years)
        tensor([1.3709e-06, 5.9605e-08, 3.4954e-01, 1.5438e-05])
        >>> for t in torch.tensor([100.0, 150.0]): survival_function(log_params, time=t)  # Subject-specific survival at multiple new times
        tensor([0.0000, 0.0000, 0.0288, 0.0000])
        tensor([0.0000, 0.0000, 0.0123, 0.0000])


    """
    log_scale, log_shape = _check_log_shape(log_params).unbind(1)

    if time.dim() == 0:
        # Use one time for each sample
        time = time.repeat(len(log_params))
    elif all([time.size(0) == log_params.size(0), all_times]):
        # Use all times for each sample
        time = time.unsqueeze(0).expand(len(time), len(time))  # expand across rows
        log_scale = log_scale.unsqueeze(1).expand(
            len(time), len(time)
        )  # expand across columns
        log_shape = log_shape.unsqueeze(1).expand(
            len(time), len(time)
        )  # expand across columns
    if time.size(0) != log_params.size(0):
        raise ValueError(
            f"Dimension mismatch: 'time' ({len(time)}) does not match the length of 'log_params' ({len(log_params)})."
        )
    return 1 - torch.distributions.weibull.Weibull(
        torch.exp(log_scale), torch.exp(log_shape)
    ).cdf(time)


def log_hazard(
    log_params: torch.Tensor, time: torch.Tensor, all_times: bool = True
) -> torch.Tensor:
    """Log hazard of the Weibull Accelerated Time Failure (AFT) survival model.

    Args:
        log_params (torch.Tensor, float):
            Parameters of the Weibull distribution of shape = (n_samples, 1) or (n_samples, 2).
            The first column corresponds to the log scale parameter. The second column
            corresponds to the log shape parameter. If the log shape parameter is missing, it is
            imputed with 0.
        time (torch.Tensor, float):
            Time at which to evaluate the log hazard.
            Should be of length n_samples to evaluate the log hazard at observed time-to-event or censoring,
            or of length one to evaluate the log hazard at a new time.
        all_times (bool):
            If True, subject-specific log hazard is evaluated at all ``time`` (used for evaluation metrics).
            If False, subject-specific log hazard is evaluated at respective ``time``.
            Defaults is True.
            Ignored if ``time`` is of length one.

    Returns:
        (torch.Tensor, float): Subject-specific log hazard evaluated at ``time``.

    Examples:
        >>> _ = torch.manual_seed(42)
        >>> time = torch.randint(low=1, high=100, size=(4,))
        >>> log_params = torch.randn((4, 2))
        >>> log_hazard(log_params, time, all_times = False)  # Log hazard at respective time
        tensor([ 0.4392, -0.0303, -3.9672,  0.9140])
        >>> log_hazard(log_params, time, all_times = True)  # Default. Log hazard at all time
        tensor([[ 0.4392,  1.1174,  1.1227,  0.9913],
                [ 0.4148, -0.0303, -0.0338,  0.0525],
                [-2.7225, -3.9575, -3.9672, -3.7279],
                [ 0.2606,  1.0632,  1.0695,  0.9140]])
        >>> log_hazard(log_params, time=torch.tensor(10.0))  # Log hazard at one new time (e.g., 10 years)
        tensor([ 0.5316,  0.3542, -2.8907,  0.3699])
        >>> for t in torch.tensor([100.0, 150.0]): log_hazard(log_params, time=t)  # Subject-specific log hazard at multiple new times
        tensor([ 1.1280, -0.0372, -3.9767,  1.0757])
        tensor([ 1.2330, -0.1062, -4.1680,  1.1999])
        >>> log_params  *= 1e2  # Increase scale
        >>> log_hazard(log_params, time, all_times = False)  # Check for Torch.Inf values
        tensor([-1.0000e+10, -2.3197e+01, -6.8385e+01, -1.0000e+10])
    """

    log_scale, log_shape = _check_log_shape(log_params).unbind(1)

    if time.dim() == 0:
        # Use fixed time for each sample
        time = time.repeat(len(log_params))
    elif all([time.size(0) == log_params.size(0), all_times]):
        # Use all times for each sample
        time = time.unsqueeze(0).expand(len(time), len(time))  # expand across rows
        log_scale = log_scale.unsqueeze(1).expand(
            len(time), len(time)
        )  # expand across columns
        log_shape = log_shape.unsqueeze(1).expand(
            len(time), len(time)
        )  # expand across columns
    if time.size(0) != log_params.size(0):
        raise ValueError(
            f"Dimension mismatch: 'time' ({len(time)}) does not match the length of 'log_params' ({len(log_params)})."
        )

    return torch.clamp(
        log_shape
        - log_scale
        + torch.expm1(log_shape)
        * (torch.log(torch.clip(time, 1e-100, torch.inf)) - log_scale),
        min=-TORCH_CLAMP_VALUE,
        max=TORCH_CLAMP_VALUE,
    )


def cumulative_hazard(
    log_params: torch.Tensor, time: torch.Tensor, all_times: bool = True
) -> torch.Tensor:
    """Cumulative hazard for the Weibull Accelerated Time Failure (AFT) survival model.

    Args:
        log_params (torch.Tensor, float):
            Parameters of the Weibull distribution of shape = (n_samples, 1) or (n_samples, 2).
            The first column corresponds to the log scale parameter. The second column
            corresponds to the log shape parameter. If the log shape parameter is missing, it is
            imputed with 0.
        time (torch.Tensor, float):
            Time-to-event or censoring of length n_samples.
        all_times (bool)
            If True, subject-specific cumulative hazard is evaluated at all ``time`` (used for evaluation metrics).
            If False, subject-specific cumulative hazard is evaluated at respective ``time``.
            Defaults is True.

    Returns:
        (torch.Tensor, float): Subject-specific cumulative hazard evaluated at ``time``.

    Examples:
        >>> _ = torch.manual_seed(42)
        >>> time = torch.randint(low=1, high=100, size=(4,))
        >>> log_params = torch.randn((4, 2))
        >>> cumulative_hazard(log_params, time, all_times=False) # Cumulative hazard at respective time
        tensor([  8.6257, 112.2115,   3.5105, 112.6339])
        >>> cumulative_hazard(log_params, time, all_times=True) # Default. Cumulative hazard at all time
        tensor([[  8.6257, 233.0865, 239.2167, 126.2805],
                [ 12.7698, 112.2115, 114.1484,  74.9134],
                [  0.8706,   3.4725,   3.5105,   2.6850],
                [  6.9530, 212.7592, 218.5687, 112.6339]])
    """
    log_scale, log_shape = _check_log_shape(log_params).unbind(1)

    if all_times:
        # Use all times for each sample
        time = time.unsqueeze(0).expand(len(time), len(time))  # expand across rows
        log_scale = log_scale.unsqueeze(1).expand(
            len(time), len(time)
        )  # expand across columns
        log_shape = log_shape.unsqueeze(1).expand(
            len(time), len(time)
        )  # expand across columns

    return torch.clamp(
        torch.exp(
            torch.exp(log_shape)
            * (torch.log(torch.clip(time, 1e-100, torch.inf)) - log_scale)
        ),
        min=0,
        max=TORCH_CLAMP_VALUE,
    )


def _check_log_shape(log_params: torch.Tensor) -> torch.Tensor:
    """Private function, check if the log shape is missing and impute it with 0
    if needed."""
    if any(
        [
            log_params.dim() == 0,
            log_params.dim() == 1,  # if shape = [n_samples]
            log_params.dim() > 1
            and log_params.size(1) == 1,  # if shape = [n_samples, 1]
        ]
    ):
        if log_params.dim() == 1:
            log_params = log_params.unsqueeze(1)

        # Missing log shape parameter. Creating zeros placeholder instead.
        log_params = torch.hstack((log_params, torch.zeros_like(log_params)))

    return log_params


def _check_inputs(log_params: torch.Tensor, event: torch.Tensor, time: torch.Tensor):
    """Private function, perform input format checks."""
    if not isinstance(log_params, torch.Tensor):
        raise TypeError("Input 'log_params' must be a tensor.")

    if not isinstance(event, torch.Tensor):
        raise TypeError("Input 'event' must be a tensor.")

    if not isinstance(time, torch.Tensor):
        raise TypeError("``Input 'time' must be a tensor.")

    if log_params.shape[0] != len(event):
        raise ValueError(
            "Length mismatch: The length of 'log_params' must match the length of 'event'."
        )

    if len(time) != len(event):
        raise ValueError(
            "Length mismatch: The length of 'time' must match the length of 'event'.`"
        )

    if any(val < 0 for val in time):
        raise ValueError("All elements in 'time' must be non-negative.")

    if any(val not in [True, False, 0, 1] for val in event):
        raise ValueError("All elements in 'event' must be boolean (True/False or 0/1).")


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



"""

Tests for torch.compile

References:
    - https://pytorch.org/tutorials/intermediate/torch_compile_tutorial.html
    - https://github.com/pytorch/pytorch/issues/122094

"""

# global modules
import unittest

import torch

# Local modules
from torchsurv.loss.cox import neg_partial_log_likelihood as cox
from torchsurv.loss.weibull import neg_log_likelihood as weibull

# set seed for reproducibility
torch.manual_seed(42)

N = 512


class TestTorchCompile(unittest.TestCase):
    """
    Tests using torch.compile with cox
    """

    def test_cox_equivalence(self):
        """
        whether the compiled version of cox evaluates to the same value
        """

        # random data and parameters
        log_hz = torch.randn(N)
        event = torch.randint(low=0, high=2, size=(N,)).bool()
        time = torch.randint(low=1, high=100, size=(N,))

        # compiled version of cox
        ccox = torch.compile(cox)

        loss_cox = cox(log_hz, event, time)
        loss_ccox = ccox(log_hz, event, time)

        self.assertTrue(torch.allclose(loss_cox, loss_ccox, rtol=1e-3, atol=1e-3))

    def test_weibull_equivalence(self):
        """
        whether the compiled version of weibull evaluates to the same value
        """

        # random data and parameters
        log_hz = torch.randn(N)
        event = torch.randint(low=0, high=2, size=(N,)).bool()
        time = torch.randint(low=1, high=100, size=(N,))

        # compiled version of weibull
        cweibull = torch.compile(weibull)

        loss_weibull = weibull(log_hz, event, time)
        loss_cweibull = cweibull(log_hz, event, time)

        self.assertTrue(
            torch.allclose(loss_weibull, loss_cweibull, rtol=1e-3, atol=1e-3)
        )


if __name__ == "__main__":

    unittest.main()



# global modules
import json
import unittest

import numpy as np
import torch
from sksurv.metrics import brier_score as brier_score_sksurv
from sksurv.metrics import integrated_brier_score as integrated_brier_score_sksurv
from utils import DataBatchContainer, conditions_ci, conditions_p_value

# Local modules
from torchsurv.metrics.brier_score import BrierScore
from torchsurv.stats.ipcw import get_ipcw

torch.manual_seed(23)

# Load the benchmark metrics from R
with open("tests/benchmark_data/benchmark_brier_score.json", "r") as file:
    benchmark_brier_scores = json.load(file)

brier_score = BrierScore()


class TestBrierScore(unittest.TestCase):
    """
    List of packages compared
        - survMetrics (R)
        - survAUC (R)
        - sksurv (Python)
    """

    def test_brier_score_real_data(self):
        """test point estimate of brier score and integrated brier score on lung and gbsg datasets"""
        for benchmark_brier_score in benchmark_brier_scores:
            train_event = torch.tensor(benchmark_brier_score["train_status"]).bool()
            train_time = torch.tensor(
                benchmark_brier_score["train_time"], dtype=torch.float32
            )
            test_event = torch.tensor(benchmark_brier_score["test_status"]).bool()
            test_time = torch.tensor(
                benchmark_brier_score["test_time"], dtype=torch.float32
            )
            estimate = torch.tensor(
                benchmark_brier_score["estimate"], dtype=torch.float32
            )
            new_time = torch.tensor(benchmark_brier_score["times"], dtype=torch.float32)

            # ipcw obtained using censoring distribution estimated on train data
            ipcw = get_ipcw(train_event, train_time, test_time)
            ipcw_new_time = get_ipcw(train_event, train_time, new_time)

            bs = brier_score(
                estimate,
                test_event,
                test_time,
                new_time=new_time,
                weight=ipcw,
                weight_new_time=ipcw_new_time,
            )
            ibs = brier_score.integral()

            bs_survAUC = np.array(benchmark_brier_score["brier_score_survAUC"])
            ibs_survAUC = np.array(benchmark_brier_score["ibrier_score_survAUC"])

            # commented out: values far off compared to other packages
            # self.assertTrue(np.allclose(bs.numpy(), bs_survAUC, rtol=1e-5, atol=1e-8))
            # self.assertTrue(np.allclose(ibs.numpy(), ibs_survAUC, rtol=1e-5, atol=1e-8))

            # ipcw obtained using censoring distribution estimated on test data
            ipcw = get_ipcw(test_event, test_time)
            ipcw_new_time = get_ipcw(test_event, test_time, new_time)

            bs = brier_score(
                estimate,
                test_event,
                test_time,
                new_time=new_time,
                weight=ipcw,
                weight_new_time=ipcw_new_time,
            )
            ibs = brier_score.integral()

            bs_survMetrics = np.array(benchmark_brier_score["brier_score_survMetrics"])
            ibs_survMetrics = np.array(
                benchmark_brier_score["ibrier_score_survMetrics"]
            )

            self.assertTrue(
                np.allclose(bs.numpy(), bs_survMetrics, rtol=1e-2, atol=1e-3)
            )
            self.assertTrue(
                np.allclose(ibs.numpy(), ibs_survMetrics, rtol=1e-2, atol=1e-3)
            )

    def test_brier_score_simulated_data(self):
        """test point estimate of brier score and integrated brier score on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "test_ties_time_event",
                "train_ties_time_censoring",
                "test_ties_time_censoring",
                "train_ties_time_event_censoring",
                "test_ties_time_event_censoring",
                "test_no_censoring",
                # "train_no_censoring", sksurv fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for batch in batch_container.batches:
            (
                train_time,
                train_event,
                test_time,
                test_event,
                _,
                new_time,
                y_train_array,
                y_test_array,
                _,
                new_time_array,
            ) = batch

            estimate = torch.rand((len(test_time), len(new_time)))

            ipcw = get_ipcw(train_event, train_time, test_time)
            ipcw_new_time = get_ipcw(train_event, train_time, new_time)
            bs = brier_score(
                estimate,
                test_event,
                test_time,
                new_time=new_time,
                weight=ipcw,
                weight_new_time=ipcw_new_time,
            )

            _, bs_sksurv = brier_score_sksurv(
                y_train_array, y_test_array, estimate.numpy(), new_time_array
            )

            self.assertTrue(np.allclose(bs.numpy(), bs_sksurv, rtol=1e-5, atol=1e-8))

            if len(new_time) > 2:
                ibs = brier_score.integral()
                ibs_sksurv = integrated_brier_score_sksurv(
                    y_train_array, y_test_array, estimate.numpy(), new_time_array
                )
                self.assertTrue(
                    np.allclose(ibs.numpy(), ibs_sksurv, rtol=1e-5, atol=1e-8)
                )

    def test_brier_score_confidence_interval_pvalue(self):
        """test brier score confidence interval and p value are as expected"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "test_ties_time_event",
                "test_ties_time_censoring",
                "test_ties_time_event_censoring",
                #'test_no_censoring', concordance_index_censored fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for batch in batch_container.batches:
            (_, _, time, event, _, new_time, *_) = batch

            estimate = torch.rand((len(time), len(new_time)))

            # Run c-index
            brier_score(estimate, event, time, new_time)

            n_bootstraps = 9

            for method in ["parametric", "bootstrap"]:
                for alternative in ["two_sided", "less", "greater"]:
                    brier_score_ci = brier_score.confidence_interval(
                        method=method,
                        alternative=alternative,
                        n_bootstraps=n_bootstraps,
                    )
                    self.assertTrue(
                        all(
                            [
                                conditions_ci(brier_score_ci[:, i])
                                for i in range(len(brier_score.brier_score))
                            ]
                        )
                    )

            for alternative in ["two_sided", "less", "greater"]:
                brier_score_pvalue = brier_score.p_value(
                    method="bootstrap",
                    alternative=alternative,
                    n_bootstraps=n_bootstraps,
                )
                self.assertTrue(
                    all(
                        [
                            conditions_p_value(brier_score_pvalue[i])
                            for i in range(len(brier_score.brier_score))
                        ]
                    )
                )
                brier_score_pvalue = brier_score.p_value(
                    method="parametric",
                    alternative=alternative,
                    n_bootstraps=n_bootstraps,
                    null_value=0.3,
                )
                self.assertTrue(
                    all(
                        [
                            conditions_p_value(brier_score_pvalue[i])
                            for i in range(len(brier_score.brier_score))
                        ]
                    )
                )

    def test_brier_score_compare(self):
        """test compare function of brier score behaves as expected"""

        _ = torch.manual_seed(42)
        n = 128
        m = 100
        estimate_informative = torch.rand((n,))  # estimate used to define time-to-event
        estimate_non_informative = (
            torch.rand((n,)).unsqueeze(1).expand(n, n)
        )  # random estimate
        event = torch.randint(low=0, high=2, size=(n,)).bool()
        time = torch.randn(size=(n,)) + estimate_informative * 20.0 + 200

        estimate_informative = estimate_informative.unsqueeze(1).expand(n, n)

        mask = event & (time < torch.max(time))
        new_time, inverse_indices, counts = torch.unique(
            time[mask], sorted=True, return_inverse=True, return_counts=True
        )
        sorted_unique_indices = BrierScore._find_torch_unique_indices(
            inverse_indices, counts
        )
        estimate_informative = (estimate_informative[:, mask])[:, sorted_unique_indices]
        estimate_non_informative = (estimate_non_informative[:, mask])[
            :, sorted_unique_indices
        ]

        ipcw = get_ipcw(event, time)  # ipcw weights at subject event time
        ipcw_new_time = get_ipcw(event, time, new_time)  # ipcw weights at new_time

        brier_score_informative = BrierScore()
        bs_informative = brier_score_informative(
            estimate_informative, event, time, new_time, ipcw, ipcw_new_time
        )[0:-1]
        ibs_informative = brier_score_informative.integral()

        brier_score_non_informative = BrierScore()
        bs_non_informative = brier_score_non_informative(
            estimate_non_informative, event, time, new_time, ipcw, ipcw_new_time
        )[0:-1]
        ibs_non_informative = brier_score_non_informative.integral()

        p_value_compare_informative = brier_score_informative.compare(
            brier_score_non_informative
        )
        p_value_compare_non_informative = brier_score_non_informative.compare(
            brier_score_informative
        )

        self.assertTrue(np.all(bs_informative.numpy() < bs_non_informative.numpy()))
        self.assertTrue(np.all(ibs_informative.numpy() < ibs_non_informative.numpy()))
        self.assertTrue(np.any(p_value_compare_informative.numpy() < 0.05))
        self.assertTrue(np.all(p_value_compare_non_informative.numpy() > 0.05))


if __name__ == "__main__":
    unittest.main()



# global modules
import json
import unittest

import numpy as np
import torch
from sksurv.metrics import concordance_index_censored, concordance_index_ipcw
from utils import DataBatchContainer, conditions_ci, conditions_p_value

# Local modules
from torchsurv.metrics.cindex import ConcordanceIndex
from torchsurv.stats.ipcw import get_ipcw

torch.manual_seed(23)

# Load the benchmark metrics from R
with open("tests/benchmark_data/benchmark_cindex.json", "r") as file:
    benchmark_cindexs = json.load(file)


cindex = ConcordanceIndex()


class TestCIndex(unittest.TestCase):
    """
    List of packages compared
        - survival (R): Harrell's C index
        - survcomp (R): Harrell's C index
        - survAUC (R): Uno's C-index
        - survC1 (R): Uno's C-index
        - sksurv (Python): Harrell's C index and Uno's C-index
    """

    def test_cindex_real_data(self):
        """test point estimate concordance index on lung and gbsg datasets"""
        for benchmark_cindex in benchmark_cindexs:
            train_event = torch.tensor(benchmark_cindex["train_status"]).bool()
            train_time = torch.tensor(
                benchmark_cindex["train_time"], dtype=torch.float32
            )
            test_event = torch.tensor(benchmark_cindex["test_status"]).bool()
            test_time = torch.tensor(benchmark_cindex["test_time"], dtype=torch.float32)
            estimate = torch.tensor(benchmark_cindex["estimate"], dtype=torch.float32)
            new_time = torch.tensor(benchmark_cindex["times"], dtype=torch.float32)

            # harrell's c-index
            c_harrell = cindex(
                estimate,
                test_event,
                test_time,
            ).numpy()

            c_harrell_survival = np.array(
                benchmark_cindex["c_Harrell_survival"]
            )  # survival
            c_harrell_survcomp = np.array(
                benchmark_cindex["c_Harrell_survcomp"]
            )  # survcomp

            # uno's c-index
            ipcw = get_ipcw(train_event, train_time, test_time)
            c_uno = cindex(
                estimate, test_event, test_time, weight=ipcw, tmax=new_time[-1]
            )

            c_uno_survAUC = benchmark_cindex["c_Uno_survAUC"]  # survAUC
            c_uno_survC1 = benchmark_cindex["c_Uno_survC1"]  # survC1

            self.assertTrue(
                np.isclose(
                    c_harrell,
                    c_harrell_survival,
                    rtol=1e-4,
                    atol=1e-8,
                )
            )

            self.assertTrue(
                np.isclose(
                    c_harrell,
                    c_harrell_survcomp,
                    rtol=1e-2,
                    atol=1e-8,
                )
            )

            self.assertTrue(
                np.isclose(c_uno.numpy(), np.array(c_uno_survAUC), rtol=1e-1, atol=1e-8)
            )

            self.assertTrue(
                np.isclose(c_uno.numpy(), np.array(c_uno_survC1), rtol=1e-1, atol=1e-8)
            )

    def test_cindex_se_real_data(self):
        """test standard error of concordance index on lung and gbsg datasets"""
        for benchmark_cindex in benchmark_cindexs:
            concordant = torch.tensor(benchmark_cindex["ch_survcomp"], dtype=torch.int)
            discordant = torch.tensor(benchmark_cindex["dh_survcomp"], dtype=torch.int)
            weight = torch.tensor(
                benchmark_cindex["weights_survcomp"], dtype=torch.float32
            )
            c_harrell_survcomp = torch.tensor(
                benchmark_cindex["c_Harrell_survcomp"], dtype=torch.float32
            )

            # overwrite objects used to calculate standard error
            cindex.concordant = concordant
            cindex.discordant = discordant
            cindex.weight = weight
            cindex.cindex = c_harrell_survcomp

            # Noether standard error
            cindex_se = cindex._concordance_index_se()
            cindex_se_survcomp = np.array(
                benchmark_cindex["c_se_noether_survcomp"]
            )  # survcomp

            # conservative confidence interval
            cindex_lower = cindex._confidence_interval_conservative(
                alpha=0.05, alternative="two_sided"
            )[0]
            cindex_lower_survcomp = np.array(
                benchmark_cindex["c_lower_conservative_survcomp"]
            )  # survcomp

            self.assertTrue(
                np.isclose(
                    cindex_se.numpy(),
                    np.array(cindex_se_survcomp),
                    rtol=1e-1,
                    atol=1e-8,
                )
            )

            self.assertTrue(
                np.isclose(
                    cindex_lower.numpy(),
                    np.array(cindex_lower_survcomp),
                    rtol=1e-1,
                    atol=1e-8,
                )
            )

    def test_cindex_simulated_data(self):
        """test estimate of concordance index on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "test_ties_time_event",
                "train_ties_time_censoring",
                # "test_ties_time_censoring",
                "train_ties_time_event_censoring",
                "test_ties_time_event_censoring",
                "test_no_censoring",
                #'train_no_censoring', concordance_index_ipcw from sksurv fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for i, batch in enumerate(batch_container.batches):
            (
                train_time,
                train_event,
                test_time,
                test_event,
                estimate,
                new_time,
                y_train_array,
                y_test_array,
                estimate_array,
                new_time_array,
            ) = batch

            # harrell's c index
            c_harrell = cindex(
                estimate,
                test_event,
                test_time,
            )
            c_harrell_sksurv = concordance_index_censored(
                y_test_array["survival"], y_test_array["futime"], estimate_array
            )[0]

            # uno's c-index
            ipcw = get_ipcw(train_event, train_time, test_time)
            c_uno = cindex(
                estimate, test_event, test_time, weight=ipcw, tmax=new_time[-1]
            ).numpy()

            c_uno_sksurv = concordance_index_ipcw(
                y_train_array, y_test_array, estimate_array, tau=new_time_array[-1]
            )[0]

            self.assertTrue(
                np.isclose(c_harrell.numpy(), c_harrell_sksurv, rtol=1e-2, atol=1e-8)
            )
            self.assertTrue(np.isclose(c_uno, c_uno_sksurv, rtol=1e-2, atol=1e-8))

    def test_cindex_confidence_interval_pvalue(self):
        """test concordance index confidence interval and p value are as expected"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "test_ties_time_event",
                "test_ties_time_censoring",
                "test_ties_time_event_censoring",
                #'test_no_censoring', concordance_index_censored fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for batch in batch_container.batches:
            (_, _, time, event, estimate, *_) = batch

            # Run c-index
            cindex(
                estimate,
                event,
                time,
            )

            n_bootstraps = 9

            for method in ["noether", "conservative", "bootstrap"]:
                for alternative in ["two_sided", "less", "greater"]:
                    self.assertTrue(
                        conditions_ci(
                            cindex.confidence_interval(
                                method=method,
                                alternative=alternative,
                                n_bootstraps=n_bootstraps,
                            )
                        )
                    )

            for method in ["noether", "bootstrap"]:
                for alternative in ["two_sided", "less", "greater"]:
                    self.assertTrue(
                        conditions_p_value(
                            cindex.p_value(
                                method=method,
                                alternative=alternative,
                                n_bootstraps=n_bootstraps,
                            )
                        )
                    )

    def test_cindex_compare(self):
        "test compare function of cindex behaves as expected"

        _ = torch.manual_seed(42)
        n = 128
        estimate_informative = torch.randn(
            (n,)
        )  # estimate used to define time-to-event
        estimate_non_informative = torch.randn((n,))  # random estimate
        event = torch.randint(low=0, high=2, size=(n,)).bool()
        time = (
            torch.randn(size=(n,)) * 10 - estimate_informative * 5.0 + 200
        )  # + estimate for cindex < 0.5 and - for cindex > 0.5

        cindex_informative = ConcordanceIndex()
        c1 = cindex_informative(estimate_informative, event, time)

        cindex_non_informative = ConcordanceIndex()
        c2 = cindex_non_informative(estimate_non_informative, event, time)

        p_value_compare = cindex_informative.compare(cindex_non_informative)

        self.assertTrue(c1.numpy() > c2.numpy())
        self.assertTrue(p_value_compare < 0.05)

    def test_cindex_error_raised(self):
        """test that error are raised in not-accepted edge cases."""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=1,
            flags_to_set=["test_all_censored"],
        )
        for batch in batch_container.batches:
            (_, _, test_time, test_event, estimate, *_) = batch

            self.assertRaises(ValueError, cindex, estimate, test_event, test_time)


if __name__ == "__main__":
    unittest.main()
